//----------------------------------------------------------------------------
// (c) Copyright 2008 Actel Corporation
//
// name:		ClosedLoopTrimDemo.v
// function:	Top-level Fusion design using Core8051s and several peripherals,
// author:      Stephane Gauthier
//
// history:		March 2008  - SNG created
//
//
//IMPORTANT-READ THESE TERMS CAREFULLY BEFORE UTILIZING THE INFORMATION CONTAINED IN THIS APPLICATION NOTE. 
//
//NO SUPPORT/NO WARRANTY.  
//THE DESIGN FILES AND DOCUMENTATION ARE PROVIDED WITHOUT SUPPORT OR WARRANTY OF ANY KIND.  
//ACTEL IS NOT OBLIGATED TO PROVIDE UPDATES, BUG FIXES, OR TECHNICAL SUPPORT AND DISCLAIMS ALL WARRANTIES INCLUDING, 
//WITHOUT LIMITATION, THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND WARRANTIES OF 
//NON-INFRINGEMENT OF THE RIGHTS OF THIRD PARTIES (INCLUDING, WITHOUT LIMITATION, RIGHTS UNDER PATENT, COPYRIGHT, 
//TRADE SECRET, OR OTHER INTELLECTUAL PROPERTY RIGHTS).  RECIPIENT ACCEPTS THE APPLICATION NOTE DESIGN FILES AND 
//DOCUMENTATION IN "AS-IS" CONDITION.
//----------------------------------------------------------------------------

module ClosedLoopTrimDemo (
    GL_RESETn,
    RS232_TX0,
    RS232_RX0,
    DACOUT,
    VIN,
    VAREF
    );

input			
    GL_RESETn,	// Active-low board pushbutton/power-on reset
	RS232_RX0,	// RS232 Rx input
    VIN;        // Voltage Rail Input

output		
    DACOUT,     // DAC output to feed low pass filter
    RS232_TX0;	// RS232 Tx output

inout
    VAREF;      // VAREF signal needs to be declared even when using int ref


//----------------------------------------------------------------------------
// Signal Declarations
//----------------------------------------------------------------------------
// Program Memory interface name directions relative to Core8051s
wire	[7:0]	memdatai;
wire	[7:0]	memdatao;
wire	[15:0]	memaddr;
wire			memrd;
wire			memwr;
wire			mempsrd;
wire            memsacki;
wire[15:0]      GP_OUT;

// Flash Memory signals
wire	[17:0]	fm_addr;
wire	[31:0]	fm_din;
wire	[31:0]	fm_dout;
wire	[1:0]	fm_byte_width;
wire			fm_busy;

wire			mempsacki_pre;
wire	[7:0]	int_mem_dout;
wire			fm_progmem;

// Initialization Signals
wire[8:0]       INIT_ADDR;
wire[8:0]       INIT_DATA;
wire            INIT_DONE;
wire            INIT_CALIBROM_WEN;
wire            INIT_CALIBCOEFF_WEN;

// Analog Interface Signals
wire[11:0]      ADC_RESULT;

// Internal signals
wire			rcosc_100mhz;
wire			proc_clk;
wire            DAC_clk;
wire            NVM_clk;
wire			proc_rstn;
wire			DAC_rstn;
reg             DAC_rstn1;
reg             DAC_rstn2;
reg[11:0]       DAC_reg;
reg             edge1;
reg             edge3;
reg             edge2;
reg             DAC_update;
wire            trigger_sample1;
reg             trigger_sample2;
reg             trigger_sample;
reg             INIT_DONE_sync;

//----------------------------------------------------------------------------
// Clocks and Resets
//----------------------------------------------------------------------------

// Instantiate internal RC Oscillator with 100MHz output
InternalOscillator u_RCOSC_100MHz (.CLKOUT(rcosc_100mhz));

// Instantiate PLL to generate clocks.
// Three clock systems were chosen so that the user could modify the DAC clock
// without affecting the processor clock.
//    - proc_clk @ 9.216 MHz to simplify UART BAUD generator
//       Also used as ACM clock (init clock) so must keep below 10MHz
//       The 8051 can easily go beyond 10MHz.  If you need higher performance
//       please contact your local Actel FAE.
//    - DAC_clk @ 18.432, but can be modified by user.  Should be able to get 
//       up to 100MHz.  Check the timing report after place and route.
//    - NVM_clk @ 4x processor clock simply to reduce the number of wait states
//       when we cross a page clock boundary
pll_main u_pll_main (
	.POWERDOWN(1'b1),
	.CLKA(rcosc_100mhz),
	.OADIVRST(1'b1),
	.LOCK(),
	.GLA(DAC_clk),
	.GLB(proc_clk),
    .GLC(NVM_clk)
);

// synchronize DAC_rstn
always @(posedge DAC_clk or negedge GL_RESETn)
    if (~GL_RESETn)
        begin
        DAC_rstn1 <= 0;
        DAC_rstn2 <= 0;
        end
    else
        begin
        DAC_rstn1 <= 1'b1;
        DAC_rstn2 <= DAC_rstn1;
        end

// Put DAC_rstn on a global
CLKINT DAC_rstn_gl(
    .A(DAC_rstn2),
    .Y(DAC_rstn)
    );

// Synchronize proc_rstn
// The CoreConsole generated system already synchronizes the reset and
// outputs it on the top level port named RESETN.

// NVM_clk is an even multiple of proc_clk, so no need to synchronize.  Just 
// use the proc_rstn for the NVM.

// We do need to synchronize the INIT_DONE signal which releases the NVM 
// for the processor.
always @(posedge proc_clk)
    INIT_DONE_sync <= INIT_DONE;

//----------------------------------------------------------------------------
// CoreDAC
//----------------------------------------------------------------------------
LowRippleDAC DAC_inst(
    .clk    (DAC_clk),
    .rstn   (DAC_rstn),
    .DACin  (DAC_reg),
    .DACout (DACOUT)
    );

// To allow the DAC to operate on a different clock, it is necessary to  
// synchronize DAC updates to proc_clk.
// Updates are on GP_OUT[11:0].  When initially writing to GP_OUT[11:0],
// GP_OUT[15] should be high, followed by a subsequent write with the 
// same values on GP_OUT[11:0] but clearing GP_OUT[15].

// Falling edge detect on GP_OUT[15]
always @(posedge DAC_clk or negedge DAC_rstn)
    if (~DAC_rstn)
        begin
        edge1 <= 0;
        edge2 <= 0;
        edge3 <= 0;
        DAC_update <= 0;
        DAC_reg <= 2048;
        end
    else
        begin
        edge1 <= GP_OUT[15];
        edge2 <= edge1;
        edge3 <= edge2;
        DAC_update <= !edge2 & edge3;
        if (DAC_update)
            DAC_reg <= GP_OUT[11:0];
        end

//----------------------------------------------------------------------------
// CoreConsole Processor Subsystem 
//----------------------------------------------------------------------------

ProcessorSystem ProcSys_inst(
    .SYSCLK(proc_clk),
    .NSYSRESET(GL_RESETn),
    .RESETN(proc_rstn),
    // 8051 
    .ExternalMemIf_MEMACKI(1'b1),
    .ExternalMemIf_MEMDATAI(memdatai),
    .ExternalMemIf_MEMPSACKI(memsacki),
    .ExternalMemIf_MEMADDR(memaddr),
    .ExternalMemIf_MEMDATAO(memdatao),
    .ExternalMemIf_MEMPSRD(mempsrd),
    .ExternalMemIf_MEMRD(memrd),
    .ExternalMemIf_MEMWR(memwr),
    // UART
    .RX(RS232_RX0),
    .TX(RS232_TX0),
    // GPIO
    .GP_OUT(GP_OUT),
    .GP_IN({2'b0,DATAVALID,ASSC_DONE,ADC_RESULT})
    );

// Writing a 0 then a 1 to GP_OUT[13] will trigger a sample
// Proc must then wait for GP_IN[12] (ASSC_DONE) before reading GP_IN[11:0] (ADC_RESULT)

assign trigger_sample1 = GP_OUT[13];

always @(posedge proc_clk)
    begin
    trigger_sample2 <= trigger_sample1;
    trigger_sample <= ~trigger_sample2 & trigger_sample1;
    end

assign memsacki = INIT_DONE_sync & ~fm_busy;

//----------------------------------------------------------------------------
// Analog Subsystem 
//----------------------------------------------------------------------------

AnalogSystem backbone_inst(
       .SYS_CLK(proc_clk),
       .SYS_RESET(proc_rstn),
       .VAREF(VAREF),
       .DATAVALID(DATAVALID),
       .VIN(VIN),
       .ASSC_SEQIN(6'b000000),
       .ASSC_XMODE(1'b1),
       .ASSC_XTRIG(trigger_sample),
       .ASSC_SEQJUMP(1'b0),
       .ASSC_SEQCHANGE(),
       .ASSC_SEQOUT(),
       .ASSC_SAMPFLAG(),
       .ASSC_DONE(ASSC_DONE),
       .ASSC_WAIT(),
       .ASSC_CHSAT(),
       .ASSC_CHLATD(),
       .INIT_DONE(INIT_DONE),
       .INIT_DATA(INIT_DATA),
       .INIT_ADDR(INIT_ADDR),
       .INIT_ACM_WEN(INIT_ACM_WEN),
       .INIT_ASSC_WEN(INIT_ASSC_WEN),
       .INIT_CALIBROM_WEN(INIT_CALIBROM_WEN),
       .INIT_CALIBCOEFF_WEN(INIT_CALIBCOEFF_WEN),
       .ADC_RESULT(ADC_RESULT),
       .ADC_CHNUMBER()
       );

//----------------------------------------------------------------------------
// Flash Memory Subsystem 
//----------------------------------------------------------------------------

// Two clients
// - AnalogSystem   : used to configure the Analog System on power-up
// - ProcessorCode  : instruction storage space for 8051
FlashMemorySystem FlashSys_inst(
        // Analog Configuration
       .INIT_CLK(proc_clk),
       .SYS_RESET(proc_rstn),
       .INIT_POWER_UP(1'b1),
       .INIT_DONE(INIT_DONE),
       .INIT_DATA(INIT_DATA),
       .INIT_ADDR(INIT_ADDR),
       .INIT_ACM_WEN(INIT_ACM_WEN),
       .INIT_ASSC_WEN(INIT_ASSC_WEN),
       .INIT_CALIBROM_WEN(INIT_CALIBROM_WEN),
       .INIT_CALIBCOEFF_WEN(INIT_CALIBCOEFF_WEN),
       // 8051 code storage
       .USER_CLK(NVM_clk),
       .USER_ADD(fm_addr),
       .USER_AUX_BLOCK(1'b0),
       .USER_DATA(fm_din),
       .USER_DISCARD_PAGE(1'b0),
       .USER_ERASE_PAGE(1'b0),
       .USER_OVERWRITE_PAGE(1'b0),
       .USER_OVERWRITE_PROT(1'b0),
       .USER_PAGELOSS_PROT(1'b0),
       .USER_PROGRAM(1'b0),
       .USER_READ(1'b1),
       .USER_READ_NEXT(1'b0),
       .USER_LOCK(1'b1),
       .USER_SPARE_PAGE(1'b0),
       .USER_UNPROT_PAGE(1'b0),
       .USER_WIDTH(fm_byte_width),
       .USER_WRITE(1'b0),
       .USER_DOUT(fm_dout),
       .USER_PAGE_STATUS(1'b0),
       .USER_DAT_VAL(fm_busy),
       .USER_NVM_STATUS()
       );

assign fm_addr			= {2'h0,memaddr[15:0]};
assign fm_byte_width	= 2'b00;
assign fm_din			= 32'b0;
assign memdatai	        = fm_dout[7:0];


endmodule

