/*
-------------------------------------------------------------------------------
-- (c) Copyright 2006 Actel Corporation
--
-- name:		debugger_design.h
-- function:	Support routines for controlling debugger_design design
-- comments:	best viewed with tabstops set to "4"
-- history:		9/24/06 - TFB created
--              10th Oct 06 - CM modified to adapt to Core8051s-based system
--
-------------------------------------------------------------------------------
*/


//------------------------------------------------------
//general typedefs
//------------------------------------------------------

typedef unsigned long uint32;
typedef unsigned int uint16;
typedef unsigned char uchar8;

typedef xdata volatile unsigned char *xvchp;

#define XBYTE ((unsigned char volatile xdata *) 0)

// --------------------------------------------------------------
// declare constants
// --------------------------------------------------------------

#define OK  0
#define ERR 1

// internal registers
#define UART2_BASE				0xF600
#define GPIO_BASE				0xF400
#define COREAI_BASE				0xF800

// CoreUARTapb registers (communicate with PC)
#define UART2_TXDATA			(XBYTE[UART2_BASE+0])
#define UART2_RXDATA			(XBYTE[UART2_BASE+4])
#define UART2_CTRL1				(XBYTE[UART2_BASE+8])
#define UART2_CTRL2				(XBYTE[UART2_BASE+12])
#define UART2_STAT				(XBYTE[UART2_BASE+16])

// CoreGPIO registers
#define GPIO_REG				(XBYTE[GPIO_BASE+0])

// CoreAI address decodes (8-bit access)
#define ACM_CTRLSTAT_LO			(XBYTE[COREAI_BASE+0x00])
#define ACM_CTRLSTAT_HI			(XBYTE[COREAI_BASE+0x01])
#define ACM_ADDR				(XBYTE[COREAI_BASE+0x02])
// UNUSED:						(XBYTE[COREAI_BASE+0x03])
#define ACM_DATA				(XBYTE[COREAI_BASE+0x04])
// UNUSED:						(XBYTE[COREAI_BASE+0x05])
#define ADC_CTRL1_LO			(XBYTE[COREAI_BASE+0x06])
#define ADC_CTRL1_HI			(XBYTE[COREAI_BASE+0x07])
#define ADC_CTRL2_LO			(XBYTE[COREAI_BASE+0x08])
#define ADC_CTRL2_HI			(XBYTE[COREAI_BASE+0x09])
#define ADC_CTRL3_LO			(XBYTE[COREAI_BASE+0x0a])
#define ADC_CTRL3_HI			(XBYTE[COREAI_BASE+0x0b])
#define ADC_CTRL4_LO			(XBYTE[COREAI_BASE+0x0c])
#define ADC_CTRL4_HI			(XBYTE[COREAI_BASE+0x0d])
#define ADC_CTRL5_LO			(XBYTE[COREAI_BASE+0x0e])
#define ADC_CTRL5_HI			(XBYTE[COREAI_BASE+0x0f])
#define ADC_STAT_LO				(XBYTE[COREAI_BASE+0x10])
#define ADC_STAT_HI				(XBYTE[COREAI_BASE+0x11])
#define RDFIFO_DOUT				(XBYTE[COREAI_BASE+0x12])
// UNUSED:						(XBYTE[COREAI_BASE+0x13])
#define RDFIFO_STAT				(XBYTE[COREAI_BASE+0x14])
// UNUSED:						(XBYTE[COREAI_BASE+0x15])
#define INT_EN_LO				(XBYTE[COREAI_BASE+0x16])
#define INT_EN_HI				(XBYTE[COREAI_BASE+0x17])
#define INT_STAT_LO				(XBYTE[COREAI_BASE+0x18])
#define INT_STAT_HI				(XBYTE[COREAI_BASE+0x19])

// CoreAI address decodes (16-bit access)
#define ACM_CTRLSTAT			(XWORD[COREAI_BASE+0x00])
#define ADC_CTRL1				(XWORD[COREAI_BASE+0x03])
#define ADC_CTRL2				(XWORD[COREAI_BASE+0x04])
#define ADC_CTRL3				(XWORD[COREAI_BASE+0x05])
#define ADC_CTRL4				(XWORD[COREAI_BASE+0x06])
#define ADC_CTRL5				(XWORD[COREAI_BASE+0x07])
#define ADC_STAT				(XWORD[COREAI_BASE+0x08])
#define INT_EN					(XWORD[COREAI_BASE+0x0b])
#define INT_STAT				(XWORD[COREAI_BASE+0x0c])

// --------------------------------------------------------------
// declare constants
// --------------------------------------------------------------

#define RX_TIMEOUT		0xff00
#define TX_TIMEOUT		0xff00

// --------------------------------------------------------------
// declare variables
// --------------------------------------------------------------


// --------------------------------------------------------------
// declare function prototypes
// --------------------------------------------------------------

// check if a character has been received - if no character has been received,
// a value of 0xffff is returned, else the actual rx character is returned
uint16 rx_char ();

// tx a character - returns 1 for success, 0 for failure
uchar8 tx_char (char c);

// tx a string
// (automatically sends CR/LF, so no '\n' needed)
void tx_str (const char * s);

// --------------------------------------------------------------
// implement functions
// --------------------------------------------------------------

// check if a character has been received - if no character has been received,
// a value of 0xffff is returned, else the actual rx character is returned
uint16 rx_char () {
	uint16 ret;
	uchar8 rxrdy; // Rx char ready status signal from CoreUART
	ret			= 0;
	rxrdy		= UART2_STAT&0x02;
	if (rxrdy) {
		ret = UART2_RXDATA;
	} else {
		ret = 0xffff;
	}
	return ret;
}

// tx a character - returns 1 for success, 0 for failure
uchar8 tx_char (char c) {
	uchar8 ret;
	uchar8 txrdy; // Tx char ready status signal from CoreUART
	uint16 i;
	ret			= 0;
	i			= 0;
	txrdy		= UART2_STAT&0x01;


	while (!txrdy) {
		txrdy		= UART2_STAT&0x01;
		// timeout, in case Tx never ready
		if (i==TX_TIMEOUT) {
			break;
		} else {
			i++;
		}
	}
	if (txrdy) {
		UART2_TXDATA		= c;
		ret				= 1;

	}
	return ret;
}

// tx a string
// (automatically sends CR/LF, so no '\n' needed)
void tx_str (const char * s) {
	while(*s)
		tx_char(*s++);
	tx_char(0x0d); // send CR before LF for <new line>
	tx_char('\n'); // send LF (newline)
}



